@@ -1,6 +1,6 @@
 # This file is part of Tryton.  The COPYRIGHT file at the top level of
 # this repository contains the full copyright notices and license terms.
-from __future__ import unicode_literals
+
 from decimal import Decimal
 from itertools import groupby
 
@@ -11,6 +11,7 @@ except ImportError:
     Null = None
 from sql.aggregate import Sum
 
+from trytond.i18n import gettext
 from trytond.model import ModelView, ModelSQL, MatchMixin, fields, \
     sequence_ordered
 from trytond.pyson import Eval, Bool, If, Id, PYSONEncoder
@@ -19,30 +20,56 @@ from trytond.pool import Pool
 from trytond.wizard import Wizard, StateView, StateAction, Button
 from trytond.transaction import Transaction
 
-from trytond.modules.product import price_digits
+from trytond.modules.product import price_digits, round_price
 
-__all__ = ['Agent', 'Plan', 'PlanLines', 'Commission',
-    'CreateInvoice', 'CreateInvoiceAsk']
+from .exceptions import FormulaError
 
 
 class Agent(ModelSQL, ModelView):
     'Commission Agent'
     __name__ = 'commission.agent'
-    party = fields.Many2One('party.party', 'Party', required=True)
+    party = fields.Many2One('party.party', "Party", required=True,
+        context={
+            'company': Eval('company', -1),
+            },
+        depends=['company'],
+        help="The party for whom the commission is calculated.")
     type_ = fields.Selection([
             ('agent', 'Agent Of'),
             ('principal', 'Principal Of'),
             ], 'Type')
     company = fields.Many2One('company.company', 'Company', required=True)
-    plan = fields.Many2One('commission.plan', 'Plan')
-    currency = fields.Many2One('currency.currency', 'Currency',
-        states={
-            'required': Bool(Eval('plan')),
-            'invisible': ~Eval('plan'),
-            },
-        depends=['plan'])
+    plan = fields.Many2One('commission.plan', "Plan",
+        help="The plan used to calculate the commission.")
+    currency = fields.Many2One('currency.currency', "Currency", required=True)
     pending_amount = fields.Function(fields.Numeric('Pending Amount',
             digits=price_digits), 'get_pending_amount')
+    selections = fields.One2Many(
+        'commission.agent.selection', 'agent', "Selections")
+
+    @classmethod
+    def __register__(cls, module_name):
+        pool = Pool()
+        Company = pool.get('company.company')
+        sql_table = cls.__table__()
+        company = Company.__table__()
+
+        super().__register__(module_name)
+        transaction = Transaction()
+        cursor = transaction.connection.cursor()
+        table = cls.__table_handler__(module_name)
+
+        # Migration from 5.4: Ensure currency is set
+        # Don't use UPDATE FROM because SQLite does not support it
+        cursor.execute(*sql_table.update(
+                columns=[sql_table.currency],
+                values=[company.select(
+                        company.currency,
+                        where=company.id == sql_table.company)],
+                where=(sql_table.currency == Null)))
+
+        # Migration from 5.4: Add not null on currency
+        table.not_null_action('currency', 'add')
 
     @staticmethod
     def default_company():
@@ -52,6 +79,14 @@ class Agent(ModelSQL, ModelView):
     def default_type_():
         return 'agent'
 
+    @classmethod
+    def default_currency(cls):
+        pool = Pool()
+        Company = pool.get('company.company')
+        company = cls.default_company()
+        if company:
+            return Company(company).currency.id
+
     @fields.depends('company', 'currency')
     def on_change_company(self):
         if self.company and not self.currency:
@@ -90,15 +125,13 @@ class Agent(ModelSQL, ModelView):
                 where=where,
                 group_by=commission.agent)
             cursor.execute(*query)
-            amounts.update(dict(cursor.fetchall()))
-        digits = cls.pending_amount.digits
-        exp = Decimal(str(10.0 ** -digits[1]))
-        for agent_id, amount in amounts.iteritems():
+            amounts.update(dict(cursor))
+        for agent_id, amount in amounts.items():
             if amount:
                 # SQLite uses float for SUM
                 if not isinstance(amount, Decimal):
                     amount = Decimal(str(amount))
-                amounts[agent_id] = amount.quantize(exp)
+                amounts[agent_id] = round_price(amount)
         return amounts
 
     @property
@@ -109,6 +142,73 @@ class Agent(ModelSQL, ModelView):
             return self.party.account_receivable_used
 
 
+class AgentSelection(sequence_ordered(), MatchMixin, ModelSQL, ModelView):
+    "Agent Selection"
+    __name__ = 'commission.agent.selection'
+    agent = fields.Many2One('commission.agent', "Agent", required=True)
+    start_date = fields.Date(
+        "Start Date",
+        domain=[
+            If(Eval('start_date') & Eval('end_date'),
+                ('start_date', '<=', Eval('end_date')),
+                ()),
+            ],
+        depends=['end_date'],
+        help="The first date that the agent will be considered for selection.")
+    end_date = fields.Date(
+        "End Date",
+        domain=[
+            If(Eval('start_date') & Eval('end_date'),
+                ('end_date', '>=', Eval('start_date')),
+                ()),
+            ],
+        depends=['start_date'],
+        help="The last date that the agent will be considered for selection.")
+    party = fields.Many2One(
+        'party.party', "Party", ondelete='CASCADE', select=True,
+        context={
+            'company': Eval('company', -1),
+            },
+        depends=['company'])
+    company = fields.Function(fields.Many2One('company.company', "Company"),
+        'on_change_with_company', searcher='search_company')
+    employee = fields.Many2One(
+        'company.employee', "Employee", select=True,
+        domain=[
+            ('company', '=', Eval('company')),
+            ],
+        depends=['company'])
+
+    @classmethod
+    def __setup__(cls):
+        super().__setup__()
+        cls._order.insert(0, ('party', 'ASC NULLS LAST'))
+        cls._order.insert(1, ('employee', 'ASC NULLS LAST'))
+
+    @fields.depends('agent', '_parent_agent.company')
+    def on_change_with_company(self, name=None):
+        if self.agent:
+            return self.agent.company.id
+
+    @classmethod
+    def search_company(cls, name, clause):
+        return [('agent.' + clause[0],) + tuple(clause[1:])]
+
+    def match(self, pattern):
+        pool = Pool()
+        Date = pool.get('ir.date')
+
+        pattern = pattern.copy()
+        if 'company' in pattern:
+            pattern.pop('company')
+        date = pattern.pop('date', None) or Date.today()
+        if self.start_date and self.start_date > date:
+            return False
+        if self.end_date and self.end_date < date:
+            return False
+        return super().match(pattern)
+
+
 class Plan(ModelSQL, ModelView):
     'Commission Plan'
     __name__ = 'commission.plan'
@@ -120,13 +220,16 @@ class Plan(ModelSQL, ModelView):
             ('default_uom', '=', Id('product', 'uom_unit')),
             ('template.type', '=', 'service'),
             ('template.default_uom', '=', Id('product', 'uom_unit')),
-            ])
+            ],
+        help="The product that is used on the invoice lines.")
     commission_method = fields.Selection([
             ('posting', 'On Posting'),
             ('payment', 'On Payment'),
             ], 'Commission Method',
-        help='When the commission is due')
-    lines = fields.One2Many('commission.plan.line', 'plan', 'Lines')
+        help="When the commission is due.")
+    lines = fields.One2Many('commission.plan.line', 'plan', "Lines",
+        help="The formulas used to calculate the commission for different "
+        "criteria.")
 
     @staticmethod
     def default_commission_method():
@@ -141,10 +244,17 @@ class Plan(ModelSQL, ModelView):
 
     def compute(self, amount, product, pattern=None):
         'Compute commission amount for the amount'
+        def parents(categories):
+            for category in categories:
+                while category:
+                    yield category
+                    category = category.parent
+
         if pattern is None:
             pattern = {}
         if product:
-            pattern['categories'] = [c.id for c in product.categories_all]
+            pattern['categories'] = [
+                c.id for c in parents(product.categories_all)]
             pattern['product'] = product.id
         else:
             pattern['categories'] = []
@@ -159,22 +269,23 @@ class PlanLines(sequence_ordered(), ModelSQL, ModelView, MatchMixin):
     'Commission Plan Line'
     __name__ = 'commission.plan.line'
     plan = fields.Many2One('commission.plan', 'Plan', required=True,
-        ondelete='CASCADE')
+        ondelete='CASCADE',
+        help="The plan to which the line belongs.")
     category = fields.Many2One(
-        'product.category', "Category", ondelete='CASCADE')
-    product = fields.Many2One('product.product', 'Product', ondelete='CASCADE')
+        'product.category', "Category", ondelete='CASCADE',
+        help="Apply only to products in the category.")
+    product = fields.Many2One('product.product', "Product", ondelete='CASCADE',
+        help="Apply only to the product.")
     formula = fields.Char('Formula', required=True,
-        help=('Python expression that will be evaluated with:\n'
-            '- amount: the original amount'))
+        help="The python expression used to calculate the amount of "
+        "commission for the line.\n"
+        "It is evaluated with:\n"
+        "- amount: the original amount")
 
     @classmethod
     def __setup__(cls):
-        super(PlanLines, cls).__setup__()
-        cls._error_messages.update({
-                'invalid_formula': ('Invalid formula "%(formula)s" in '
-                    'commission plan line "%(line)s" with exception '
-                    '"%(exception)s".'),
-                })
+        super().__setup__()
+        cls.__access__.add('plan')
 
     @staticmethod
     def default_formula():
@@ -192,12 +303,12 @@ class PlanLines(sequence_ordered(), ModelSQL, ModelView, MatchMixin):
         try:
             if not isinstance(self.get_amount(**context), Decimal):
                 raise ValueError
-        except ValueError, exception:
-            self.raise_user_error('invalid_formula', {
-                    'formula': self.formula,
-                    'line': self.rec_name,
-                    'exception': exception,
-                    })
+        except Exception as exception:
+            raise FormulaError(
+                gettext('commission.msg_plan_line_invalid_formula',
+                    formula=self.formula,
+                    line=self.rec_name,
+                    exception=exception)) from exception
 
     def get_amount(self, **context):
         'Return amount (as Decimal)'
@@ -222,13 +333,16 @@ class Commission(ModelSQL, ModelView):
         }
     _readonly_depends = ['invoice_line']
     origin = fields.Reference('Origin', selection='get_origin', select=True,
-        readonly=True)
+        readonly=True,
+        help="The source of the commission.")
     date = fields.Date('Date', select=True, states=_readonly_states,
-        depends=_readonly_depends)
+        depends=_readonly_depends,
+        help="When the commission is due.")
     agent = fields.Many2One('commission.agent', 'Agent', required=True,
         states=_readonly_states, depends=_readonly_depends)
     product = fields.Many2One('product.product', 'Product', required=True,
-        states=_readonly_states, depends=_readonly_depends)
+        states=_readonly_states, depends=_readonly_depends,
+        help="The product that is used on the invoice line.")
     amount = fields.Numeric('Amount', required=True, digits=price_digits,
         domain=[('amount', '!=', 0)],
         states=_readonly_states, depends=_readonly_depends)
@@ -244,8 +358,11 @@ class Commission(ModelSQL, ModelView):
                 ('', ''),
                 ('invoiced', 'Invoiced'),
                 ('paid', 'Paid'),
-                ('cancel', 'Canceled'),
-                ], 'Invoice State'), 'get_invoice_state')
+                ('cancelled', 'Cancelled'),
+                ], "Invoice State",
+            help="The current state of the invoice "
+            "that the commission appears on."),
+        'get_invoice_state')
 
     @classmethod
     def __setup__(cls):
@@ -253,6 +370,7 @@ class Commission(ModelSQL, ModelView):
         cls._buttons.update({
                 'invoice': {
                     'invisible': Bool(Eval('invoice_line')),
+                    'depends': ['invoice_line'],
                     },
                 })
 
@@ -265,11 +383,9 @@ class Commission(ModelSQL, ModelView):
     def get_origin(cls):
         pool = Pool()
         Model = pool.get('ir.model')
+        get_name = Model.get_name
         models = cls._get_origin()
-        models = Model.search([
-                ('model', 'in', models),
-                ])
-        return [(None, '')] + [(m.model, m.name) for m in models]
+        return [(None, '')] + [(m, get_name(m)) for m in models]
 
     @fields.depends('agent')
     def on_change_with_currency(self, name=None):
@@ -294,17 +410,16 @@ class Commission(ModelSQL, ModelView):
         if self.invoice_line:
             state = 'invoiced'
             invoice = self.invoice_line.invoice
-            if invoice:
-                if invoice.state == 'paid':
-                    state = 'paid'
-                elif invoice.state == 'cancel':
-                    state = 'cancel'
+            if invoice and invoice.state in {'paid', 'cancelled'}:
+                state = invoice.state
         return state
 
     @classmethod
     def copy(cls, commissions, default=None):
         if default is None:
             default = {}
+        else:
+            default = default.copy()
         default.setdefault('invoice_line', None)
         return super(Commission, cls).copy(commissions, default=default)
 
@@ -314,20 +429,23 @@ class Commission(ModelSQL, ModelView):
         pool = Pool()
         Invoice = pool.get('account.invoice')
 
-        key = lambda c: c._group_to_invoice_key()
-        commissions.sort(key=key)
+        def invoice_key(c):
+            return c._group_to_invoice_key()
+
+        def line_key(c):
+            return c._group_to_invoice_line_key()
+        commissions.sort(key=invoice_key)
         invoices = []
         to_write = []
-        for key, commissions in groupby(commissions, key=key):
+        for key, commissions in groupby(commissions, key=invoice_key):
             commissions = list(commissions)
             key = dict(key)
             invoice = cls._get_invoice(key)
             invoice.save()
             invoices.append(invoice)
 
-            key = lambda c: c._group_to_invoice_line_key()
-            commissions.sort(key=key)
-            for key, commissions in groupby(commissions, key=key):
+            commissions.sort(key=line_key)
+            for key, commissions in groupby(commissions, key=line_key):
                 commissions = [c for c in commissions if not c.invoice_line]
                 key = dict(key)
                 invoice_line = cls._get_invoice_line(key, invoice, commissions)
@@ -398,8 +516,11 @@ class Commission(ModelSQL, ModelView):
 
         invoice_line = InvoiceLine()
         invoice_line.invoice = invoice
+        invoice_line.currency = invoice.currency
+        invoice_line.company = invoice.company
         invoice_line.type = 'line'
-        invoice_line.product = product
+        # Use product.id to instantiate it with the correct context
+        invoice_line.product = product.id
         invoice_line.quantity = 1
 
         invoice_line.on_change_product()
@@ -434,8 +555,10 @@ class CreateInvoice(Wizard):
     def do_create_(self, action):
         pool = Pool()
         Commission = pool.get('commission')
-        commissions = Commission.search(self.get_domain(),
-            order=[('agent', 'DESC'), ('date', 'DESC')])
+        with Transaction().set_context(_check_access=True):
+            commissions = Commission.search(self.get_domain(),
+                order=[('agent', 'DESC'), ('date', 'DESC')])
+        commissions = Commission.browse(commissions)
         Commission.invoice(commissions)
         invoice_ids = list({c.invoice_line.invoice.id for c in commissions})
         encoder = PYSONEncoder()
@@ -453,18 +576,21 @@ class CreateInvoiceAsk(ModelView):
             If(Eval('to') & Eval('from_'), [('from_', '<=', Eval('to'))],
                 []),
             ],
-        depends=['to'])
+        depends=['to'],
+        help="Limit to commissions from this date.")
     to = fields.Date('To',
         domain=[
             If(Eval('from_') & Eval('to'), [('to', '>=', Eval('from_'))],
                 []),
             ],
-        depends=['from_'])
+        depends=['from_'],
+        help="Limit to commissions to this date.")
     type_ = fields.Selection([
             ('in', 'Incoming'),
             ('out', 'Outgoing'),
             ('both', 'Both'),
-            ], 'Type')
+            ], 'Type',
+        help="Limit to commissions of this type.")
 
     @staticmethod
     def default_type_():
